#!/usr/bin/env python3
import json
import re
from argparse import ArgumentParser
from enum import Enum
from sys import stderr
from typing import List, Tuple

ReplaceMap = List[Tuple[bytes, bytes]]

PHP_SER_RE = rb"^s:(\d+):\""

MYSQL_CHARS = [
    (b"\\0", b"\0"),
    (b"\\'", b"'"),
    (b'\\"', b'"'),
    (b"\\b", b"\b"),
    (b"\\n", b"\n"),
    (b"\\r", b"\r"),
    (b"\\t", b"\t"),
    (b"\\Z", bytes([26])),
    (b"\\\\", b"\\"),
]


def parse_args():
    """
    Configure the arguments parser and parses the arguments. Returns the
    parsing result.
    """

    parser = ArgumentParser(
        description="Replaces a string by another safely even if it lies "
        "within a PHP serialized value"
    )

    parser.add_argument("search")
    parser.add_argument("replace")
    parser.add_argument("--debug", "-d", action="store_true", help="Enable debug")

    return parser.parse_args()


class StringType(Enum):
    """
    Different types of strings that can be recognized
    """

    RAW = 0
    PHP_SER = 1
    JSON = 2
    MYSQL = 3


def multi_replace(seq: bytes, mapping: ReplaceMap, reverse=False) -> bytes:
    """
    Replaces multiple search/replace couples at once
    """

    if reverse:
        d = dict((b, a) for a, b in mapping)
    else:
        d = dict(mapping)

    pattern = re.compile(b"|".join(re.escape(k) for k in d.keys()))

    return pattern.sub(lambda m: d[m.group(0)], seq)


def split(line):
    """
    Given a line of data, emit a couple (segment, type) which are the different
    kind of recognized strings
    """

    i = 0
    raw_start = 0

    while i < len(line):
        length = 0
        start = 0
        rest = line[i:]
        ser_m = re.match(PHP_SER_RE, rest)
        json_m = re.match(
            rb"\"(?:[^\"\\\0-\x1F\x7F\r\n]|\\(?:[\"\\/bfnrt]|u[a-fA-F0-9]{4}))*\"", rest
        )
        mysql_m = re.match(rb"'(?:[^'\\\r\n]|\\['\"0bnrtZ\\%_])*'", rest)

        is_php_ser = False
        is_json = bool(json_m)
        is_mysql = bool(mysql_m)

        if ser_m:
            length = int(ser_m.group(1))
            start = i + len(ser_m.group(0))
            try:
                is_php_ser = line[start + length : start + length + 2] == b'";'
            except IndexError:
                is_php_ser = False

        if is_php_ser or is_json or is_mysql:
            if raw_start != i:
                yield line[raw_start:i], StringType.RAW

        if is_php_ser:
            yield line[i : start + length + 2], StringType.PHP_SER
            i += len(ser_m.group(0)) + length + 2
            raw_start = i
        elif is_json:
            yield line[i : i + len(json_m.group(0))], StringType.JSON
            i += len(json_m.group(0))
            raw_start = i
        elif is_mysql:
            yield line[i : i + len(mysql_m.group(0))], StringType.MYSQL
            i += len(mysql_m.group(0))
            raw_start = i
        else:
            i += 1

    if i != raw_start:
        yield line[raw_start:], StringType.RAW


def uncap_json(s: bytes) -> bytes:
    return json.loads(s.decode("utf-8")).encode("utf-8")


def encapsulate_json(s: bytes) -> bytes:
    return json.dumps(s.decode("utf-8"), ensure_ascii=False).encode("utf-8")


def uncap_php_ser(s: bytes) -> bytes:
    m = re.match(PHP_SER_RE, s)
    return s[len(m.group(0)) : -2]


def encapsulate_php_ser(s: bytes) -> bytes:
    return b"s:" + f"{len(s)}".encode("utf-8") + b':"' + s + b'";'


def uncap_mysql(s: bytes) -> bytes:
    return multi_replace(s[1:-1], MYSQL_CHARS, reverse=False)


def encapsulate_mysql(s: bytes) -> bytes:
    return b"'" + multi_replace(s, MYSQL_CHARS, reverse=True) + b"'"


def uncap(s: bytes, type_: StringType) -> bytes:
    """
    Given a string literal of type_, extracts and returns its content
    """

    if type_ == StringType.PHP_SER:
        return uncap_php_ser(s)
    elif type_ == StringType.JSON:
        return uncap_json(s)
    elif type_ == StringType.MYSQL:
        return uncap_mysql(s)


def encapsulate(s: bytes, type_: StringType) -> bytes:
    """
    Encodes some content into something valid in the target string type
    """

    if type_ == StringType.PHP_SER:
        return encapsulate_php_ser(s)
    elif type_ == StringType.JSON:
        return encapsulate_json(s)
    elif type_ == StringType.MYSQL:
        return encapsulate_mysql(s)


def walk(data: bytes, mapping: ReplaceMap, depth=None) -> bytes:
    """
    Walks down the data and replaces things as it goes
    """

    out = b""

    for segment, type_ in split(data):
        if depth is not None:
            prefix = "-" * (1 + depth)
            stderr.write(f"{prefix}> {type_.name}: {segment.decode()}\n")
            stderr.flush()

        if type_ == StringType.RAW:
            out += multi_replace(segment, mapping)
        else:
            raw = uncap(segment, type_)
            replaced = walk(
                uncap(segment, type_), mapping, depth + 1 if depth is not None else None
            )

            if replaced != raw:
                out += encapsulate(replaced, type_)
            else:
                out += segment

    return out



from shlex import quote
from shutil import rmtree
from subprocess import DEVNULL, PIPE, Popen, TimeoutExpired
from tempfile import mkdtemp
from typing import Dict, Optional, Text, Tuple, Union

from luh3417.utils import make_doer

doing = make_doer("luh3417.luhssh")


def make_ssh_args(
    user: Text,
    host: Text,
    port: Union[int, Text, None] = None,
    options: Optional[Dict] = None,
    compress: bool = False,
    forward_agent: bool = False,
    nothing: bool = False,
):
    """
    Generates the appropriate SSH CLI args

    :param user: Username
    :param host: Host
    :param port: Port number (None to use default)
    :param options: SSH options (passed using -o)
    :param compress: Compress the connection
    :param forward_agent: Enable agent forwarding
    :param nothing: Just open the connection, don't run anything
    """

    out = ["ssh"]

    if options:
        for k, v in options.items():
            out.append("-o")
            out.append(f"{k}={v}")

    if port:
        out += ["-p", f"{port}"]

    if compress:
        out += ["-C"]

    if nothing:
        out += ["-N"]

    if forward_agent:
        out += ["-A"]

    return out + [f"{user}@{host}"]


class SshManager:
    """
    A manager of SSH connections. If you might open SSH connections, there is
    some cleanup to do and make sure that you eventually call shutdown().

    This allows you to multiplex SSH connections by opening a master connection
    first and then provides you SSH CLI arguments that use this multiplexed
    connection, which you can use with Popen() and so on.

    Use the following way:

    >>> try:
    >>>     manager = SshManager.instance('root', 'server.com')
    >>>     args = manager.get_args(['uptime'])
    >>>     p = Popen(args)
    >>>     p.wait()
    >>> finally:
    >>>     SshManager.shutdown()
    """

    _instances: Dict[Tuple, "SshManager"] = {}

    forward_agent = True
    compress = False

    def __init__(self, user: Text, host: Text, port: Text):
        """
        Dont call directly! Use instance() instead.
        """

        self.user: Text = user
        self.host: Text = host
        self.port: Text = port
        self.control_dir: Text = None
        self.process: Popen = None

    @property
    def control(self):
        """
        Generates the path to the control socket
        """

        return f"{self.control_dir}/control"

    def start(self):
        """
        Starting the master connection
        """

        doing.logger.debug(f"Connecting SSH to {self.user}@{self.host}")

        self.control_dir = mkdtemp()

        self.process = Popen(
            make_ssh_args(
                self.user,
                self.host,
                self.port,
                options={
                    "ControlPath": self.control,
                    "ControlMaster": "yes",
                    "ControlPersist": "no",
                },
                nothing=True,
                forward_agent=self.forward_agent,
                compress=self.compress,
            ),
            stdin=DEVNULL,
            stdout=DEVNULL,
            stderr=PIPE,
        )

        doing.logger.debug(f"Process {self.process.pid} created")

    def cleanup(self):
        """
        Cleaning up this particular connection
        """

        doing.logger.debug(f"Tearing down SSH connection to {self.user}@{self.host}")

        if self.process:
            try:
                self.process.terminate()
                self.process.communicate(timeout=10)
            except TimeoutExpired:
                self.process.kill()
                doing.logger.debug(f"Process {self.process.pid} killed")
            else:
                doing.logger.debug(f"Process {self.process.pid} terminated")

        if self.control_dir:
            rmtree(self.control_dir)

    def get_args(self, args):
        """
        Generating args for a child connection using this master
        """

        return make_ssh_args(
            self.user,
            self.host,
            self.port,
            options={"ControlPath": self.control, "ControlMaster": "no"},
            compress=self.compress,
            forward_agent=self.forward_agent,
        ) + [quote(a) for a in args]

    @classmethod
    def instance(cls, user, host, port=None) -> "SshManager":
        """
        Gets or create and start the manager instance for this user and host
        """

        if port is None:
            port = 22

        key = (user, host, port)

        if key not in cls._instances:
            cls._instances[key] = cls(user, host, port)
            cls._instances[key].start()

        return cls._instances[key]

    @classmethod
    def shutdown(cls):
        """
        Global shutdown of all manager instances
        """

        if cls._instances:
            doing.logger.debug("Shutting down SSH connections")

            to_delete = []

            for k, manager in cls._instances.items():
                manager.cleanup()
                to_delete.append(k)

            for k in to_delete:
                del cls._instances[k]



import subprocess
from dataclasses import dataclass
from subprocess import DEVNULL, PIPE
from typing import List, Optional, Text, TextIO

from luh3417.luhfs import LocalLocation, Location, SshLocation
from luh3417.luhssh import SshManager
from luh3417.serialized_replace import ReplaceMap, walk
from luh3417.utils import LuhError


def create_from_source(wp_config, source: Location):
    """
    Using a Location object and the WP config, generates the appropriate LuhSql
    object
    """

    if isinstance(source, SshLocation):
        ssh_user = source.user
        ssh_host = source.host
    elif isinstance(source, LocalLocation):
        ssh_user = None
        ssh_host = None
    else:
        raise LuhError(f"Unknown source type: {source.__class__.__name__}")

    return LuhSql(
        host=wp_config["db_host"],
        user=wp_config["db_user"],
        password=wp_config["db_password"],
        db_name=wp_config["db_name"],
        ssh_user=ssh_user,
        ssh_host=ssh_host,
    )


def create_root_from_source(wp_config, mysql_root, source: Location) -> "LuhSql":
    """
    Based on the regular DB accessor, create a root version which will be able
    to run DB manipulation queries
    """

    db = create_from_source(wp_config, source)
    method = mysql_root.get("method")
    options = mysql_root.get("options", {})

    try:
        if method == "socket":
            db.db_name = None
            db.user = options.get("mysql_user", "root")
            db.sudo_user = options.get("sudo_user")
            return db
        else:
            raise LuhError(f"Wrong MySQL root method: {method}")
    except KeyError as e:
        raise LuhError(f"Missing key for mysql_root: {e}")


def patch_sql_dump(source_path: Text, dest_path: Text, replace: ReplaceMap):
    """
    Patches the SQL dump found at source_path into a new SQL dump found in
    dest_path. It will use the replace map to replace values.

    Values are replaced in a holistic way so that PHP serialized values are
    not broken and escaped character are detected as such. This is by far not
    perfect but seems sufficient for most use cases.
    """

    try:
        with open(source_path, "rb") as i, open(dest_path, "wb") as o:
            for line in i:
                o.write(walk(line, replace))
    except OSError as e:
        raise LuhError(f"Could not open SQL dump: {e}")


@dataclass
class LuhSql:
    """
    A helper class to access MySQL locally or remotely through a SSH proxy
    """

    host: Text
    user: Text
    password: Text
    db_name: Text
    ssh_user: Optional[Text]
    ssh_host: Optional[Text]
    sudo_user: Optional[Text] = None

    def ssh_args(self, args: List[Text]) -> List[Text]:
        """
        Appends SSH connection args if required
        """

        if self.ssh_host and self.ssh_user:
            return SshManager.instance(self.ssh_user, self.ssh_host).get_args(args)
        else:
            return args

    def sudo_args(self, args: List[Text]) -> List[Text]:
        """
        Appends sudo privilege escalation args if required
        """

        if self.sudo_user:
            return ["sudo", "-u", self.sudo_user] + list(args)
        else:
            return list(args)

    def mysql_args(
        self, command: Text, extra_args: Optional[List[Text]] = None
    ) -> List[Text]:
        """
        Generates the MySQL connection arguments depending on the connection
        method and so on
        """

        out = [command] + (extra_args if extra_args else [])

        if self.password:
            out += [f"-p{self.password}"]

        out += ["-u", self.user]
        out += ["-h", self.host]

        if self.db_name:
            out += [self.db_name]

        return out

    def args(
        self, command: Text, extra_args: Optional[List[Text]] = None
    ) -> List[Text]:
        """
        Generates the proper arguments for this command and the connection
        configuration
        """

        args = self.mysql_args(command, extra_args)
        args = self.sudo_args(args)
        args = self.ssh_args(args)

        return args

    def dump_to_file(self, file_path: Text):
        """
        Dumps the database into the specified file
        """

        with open(file_path, "w", encoding="utf-8") as f:
            p = subprocess.Popen(
                self.args("mysqldump", ["--hex-blob"]),
                stderr=PIPE,
                stdout=f,
                stdin=DEVNULL,
                encoding="utf-8",
            )

            _, err = p.communicate()

            if p.returncode:
                raise LuhError(f"Could not dump MySQL DB: {err}")

    def restore_dump(self, fp: TextIO):
        """
        Restores a dump into the DB, reading the dump from an input TextIO
        (which can be the stdout of another process or simply an open file, by
        example).
        """

        p = subprocess.Popen(
            self.args("mysql"), stderr=PIPE, stdout=DEVNULL, stdin=fp, encoding="utf-8"
        )

        _, err = p.communicate()

        if p.returncode:
            raise LuhError(f"Could not import MySQL DB: {err}")

    def run_query(self, query: Text):
        """
        Runs a single SQL query
        """

        p = subprocess.Popen(
            self.args("mysql"),
            stderr=PIPE,
            stdout=DEVNULL,
            stdin=PIPE,
            encoding="utf-8",
        )

        _, err = p.communicate(query)

        if p.returncode:
            raise LuhError(f"Could not run MySQL query: {err}")



import re
import subprocess
from dataclasses import dataclass, replace
from pathlib import Path
from posixpath import join
from shlex import quote
from subprocess import CompletedProcess, Popen
from typing import Text, Tuple

from luh3417.luhssh import SshManager
from luh3417.utils import LuhError

SSH_RE = re.compile(r"^([a-zA-Z0-9_-]+)@((?:[a-zA-Z0-9-]+\.)*(?:[a-zA-Z0-9-]+)):(.*)$")


def parse_location(location: Text) -> "Location":
    """
    Guess the location type and generates the appropriate object
    """

    sm = SSH_RE.match(location)

    if sm:
        return SshLocation(user=sm.group(1), host=sm.group(2), path=sm.group(3))
    else:
        return LocalLocation(path=location)


@dataclass
class Location:
    """
    Base location object. This abstraction allows to do very specific
    operations on local or remote file systems. This is not a generic file
    system implementation but rather a very specialized one so that every
    operation can be implemented the most possibly efficient way.
    """

    path: Text

    def get_content(self) -> Text:
        """
        Calling this returns the whole content of the file
        """

        raise NotImplementedError

    def set_content(self, content) -> None:
        """
        Creates or overrides the file so it receives the provided content.
        The parent directory must exist and the location must not be a
        directory.
        """

        out, err, ret = self.run_script(
            f"echo -n {quote(content)} > {quote(self.path)}"
        )

        if ret:
            raise LuhError(f"Cannot set content: {err}")

    def ensure_exists_as_dir(self) -> None:
        """
        This ensures that the location is a directory and exists (as well as
        all parents)
        """

        raise NotImplementedError

    def archive_local_dir(self, local_path) -> None:
        """
        Puts all the content of `local_path` into a TAR/GZ archive at the
        current location.

        Beware it's probably the opposite of what you imagined (:
        """

        raise NotImplementedError

    def extract_archive_to_dir(self, target_dir: Text) -> None:
        """
        If the file at this location is an archive, then extract its content
        into the specified target_dir. Otherwise raise an error.
        """

        raise NotImplementedError

    def run_script(self, script: Text) -> Tuple[Text, Text, int]:
        """
        Runs the provided script with bash, on the machine targeted by this
        location.
        """

        raise NotImplementedError

    def chown(self, owner: Text) -> None:
        """
        Chown recursively this location
        """

        raise NotImplementedError

    def exists(self):
        """
        Returns True if this location exists
        """

        out, err, ret = self.run_script(
            f"""
                if [ ! -e {quote(self.path)} ]
                then
                    exit 1
                fi
            """
        )

        return not ret

    def set_git_repo(self, repo: Text, version: Text):
        """
        Sets the current location to be a git repo at the given version. Any
        pre-existing file or directory at this location will be overridden.
        """

        location = self.path

        if location and location[-1] == "/":
            location = location[0:-1]

        out, err, ret = self.run_script(
            """
                git clone -b {version} {repo} {location}__ \\
                && mv {location} {location}___ \\
                && mv {location}__ {location} \\
                && rm -fr {location}___
            """.format(
                repo=quote(repo), version=quote(version), location=quote(location)
            )
        )

        if ret:
            raise LuhError(f"Could not clone repo: {err}")

    def child(self, file_name) -> "Location":
        """
        Generates the location object for a child file named file_name
        """

        return replace(self, path=join(self.path, file_name))

    def rsync_path(self, as_dir: bool = True):
        """
        Generates the equivalent rsync path for this location
        """

        path = f"{self}"

        if as_dir and (not path or path[-1] != "/"):
            path += "/"
        elif not as_dir and path and path[-1] == "/":
            path = path[0:-1]

        return path


@dataclass
class SshLocation(Location):
    """
    A file located on a remote server accessed through SSH
    """

    user: Text
    host: Text
    path: Text

    def __str__(self):
        return f"{self.user}@{self.host}:{self.path}"

    @property
    def ssh_target(self):
        """
        Generates the target to give to SSH for this location
        """

        return f"{self.user}@{self.host}"

    def ssh_run(self, args, *p_args, **kwargs) -> CompletedProcess:
        """
        Runs a process remotely using subprocess.run(). This will enforce
        an UTF-8 encoding for stdin/out. Otherwise it's the same argument
        as run() and the SSH command is automatically appended to the args.
        """

        kwargs = dict(kwargs, encoding="utf-8")

        new_args = SshManager.instance(self.user, self.host).get_args(args)

        cp = subprocess.run(new_args, *p_args, **kwargs)

        if cp.returncode == 255:
            raise LuhError(
                f"SSH connection to {self.ssh_target} could not be established"
            )

        return cp

    def ssh_popen(self, args, *p_args, **kwargs) -> Popen:
        """
        Opens a process through SSH. It's the same arguments as Popen() except
        that the SSH command will be prepended to the args.
        """

        new_args = SshManager.instance(self.user, self.host).get_args(args)

        cp = subprocess.Popen(new_args, *p_args, **kwargs)

        if cp.returncode == 255:
            raise LuhError(
                f"SSH connection to {self.ssh_target} could not be established"
            )

        return cp

    def get_content(self) -> Text:
        """
        Uses a remote cat to get the content
        """

        cp = self.ssh_run(
            ["cat", self.path], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL
        )

        if cp.returncode == 1:
            raise LuhError(
                f"The file {self} does not exist or you don't have "
                f"permissions to read it"
            )
        elif cp.returncode:
            raise LuhError(f"Unknown error while reading {self}")

        return cp.stdout

    def ensure_exists_as_dir(self) -> None:
        """
        Runs mkdir -p remotely
        """

        cp = self.ssh_run(
            ["mkdir", "-p", self.path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        if cp.returncode:
            raise LuhError(f"Could not create {self} as a directory: {cp.stderr}")

    def archive_local_dir(self, local_path: Text):
        """
        Generates the archive locally and pipe it to a remote dd to write it
        on disk on the other side
        """

        tar = subprocess.Popen(
            ["tar", "-C", local_path, "-c", "-z", "."],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        dd = self.ssh_popen(
            ["dd", f"of={self.path}"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            stdin=tar.stdout,
        )

        _, dd_err = dd.communicate()
        _, tar_err = tar.communicate()

        if dd.returncode:
            raise LuhError(f"Could not write remote archive: {dd_err}")

        if tar.returncode:
            raise LuhError(f"Could not create the archive: {tar_err}")

    def extract_archive_to_dir(self, target_dir: Text) -> None:
        """
        Cat the remote file and pipe it into tar
        """

        parse_location(target_dir).ensure_exists_as_dir()

        cat = self.ssh_popen(
            ["cat", self.path], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        tar = subprocess.Popen(
            ["tar", "-C", target_dir, "-x", "-z"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            stdin=cat.stdout,
        )

        _, tar_err = tar.communicate()
        _, cat_err = cat.communicate()

        if cat.returncode:
            raise LuhError(f"Error while reading the remote archive: {cat_err}")

        if tar.returncode:
            raise LuhError(f"Error while extracting the archive: {tar_err}")

    def chown(self, owner: Text) -> None:
        """
        Simply use chown
        """

        cp = self.ssh_run(
            ["chown", "-R", owner, self.path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        if cp.returncode:
            raise LuhError(f"Failed to chown: {cp.stderr[:1000]}")

    def run_script(self, script: Text) -> Tuple[Text, Text, int]:
        """
        Runs the script through SSH piping
        """

        cp = self.ssh_popen(
            ["bash"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )

        cp.stdin.write(script)

        out, err = cp.communicate()

        return out, err, cp.returncode


@dataclass
class LocalLocation(Location):
    """
    Runs everything locally, everything inside is pretty straightforward
    """

    path: Text

    def __str__(self):
        return self.path

    def get_content(self) -> Text:
        try:
            with open(self.path, "r", encoding="utf-8") as f:
                return f.read()
        except PermissionError:
            raise LuhError(f"You don't have the permission to read {self}")
        except FileNotFoundError:
            raise LuhError(f"The file {self} does not exist")
        except OSError:
            raise LuhError(f"Unknown error while opening {self}")

    def ensure_exists_as_dir(self) -> None:
        try:
            Path(self.path).mkdir(parents=True, exist_ok=True)
        except PermissionError:
            raise LuhError(f"You don't have the permission the create {self}")
        except NotADirectoryError:
            raise LuhError(f"Some component of {self} is not a directory")
        except OSError:
            raise LuhError(f"Unknown error while creating {self}")

    def archive_local_dir(self, local_path):
        cp = subprocess.run(
            ["tar", "-C", local_path, "-c", "-z", "-f", self.path, "."],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if cp.returncode:
            raise LuhError(f"Could not create archive {self.path}")

    def extract_archive_to_dir(self, target_dir: Text) -> None:
        """
        Plain old local archive extraction
        """

        parse_location(target_dir).ensure_exists_as_dir()

        tar = subprocess.run(
            ["tar", "-C", target_dir, "-x", "-z", "-f", self.path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        if tar.returncode:
            raise LuhError(f"Error while extracting the archive: {tar.stderr}")

    def chown(self, owner: Text) -> None:
        """
        A simple local chown
        """

        cp = subprocess.run(
            ["chown", "-R", owner, self.path],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        if cp.returncode:
            raise LuhError(f"Failed to chown: {cp.stderr[:1000]}")

    def run_script(self, script: Text) -> Tuple[Text, Text, int]:
        """
        Just pipe the script to bash
        """

        cp = subprocess.Popen(
            ["bash"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )

        cp.stdin.write(script)

        out, err = cp.communicate()

        return out, err, cp.returncode



from argparse import ArgumentParser, Namespace
from os.path import join
from tempfile import TemporaryDirectory
from typing import Optional, Sequence

from luh3417.luhfs import Location, parse_location
from luh3417.luhphp import set_wp_config_values
from luh3417.luhsql import create_from_source, patch_sql_dump
from luh3417.restore import (
    configure_dns,
    ensure_db_exists,
    get_remote,
    get_wp_config,
    install_outer_files,
    make_replace_map,
    patch_config,
    read_config,
    restore_db,
    restore_files,
    run_post_install,
    run_queries,
)
from luh3417.utils import make_doer, run_main, setup_logging

doing = make_doer("luh3417.restore")


def parse_args(args: Optional[Sequence[str]] = None) -> Namespace:
    """
    Parse arguments from CLI
    """

    parser = ArgumentParser(description="Restores a snapshot")

    parser.add_argument(
        "-p", "--patch", help="A settings patch file", type=parse_location
    )
    parser.add_argument(
        "-a",
        "--allow-in-place",
        help="Allow to restore the backup in-place, overriding its origin",
        action="store_true",
    )

    parser.add_argument(
        "snapshot",
        help=(
            "Location of the snapshot file. Syntax: `~/snap.tar.gz` or "
            "`user@host:snap.tar.gz`"
        ),
        type=parse_location,
    )

    return parser.parse_args(args)


def main(args: Optional[Sequence[str]] = None):
    """
    Executes things in order
    """

    setup_logging()
    args = parse_args(args)
    snap: Location = args.snapshot

    with TemporaryDirectory() as d:
        with doing("Extracting archive"):
            snap.extract_archive_to_dir(d)

        with doing("Reading configuration"):
            config = patch_config(
                read_config(join(d, "settings.json")), args.patch, args.allow_in_place
            )

        dump = join(d, "dump.sql")

        if config["replace_in_dump"]:
            with doing("Patch the SQL dump"):
                new_dump = join(d, "dump_patched.sql")
                patch_sql_dump(
                    dump, new_dump, make_replace_map(config["replace_in_dump"])
                )
                dump = new_dump

        if config["php_define"]:
            with doing("Patch wp-config.php"):
                set_wp_config_values(
                    config["php_define"], join(d, "wordpress", "wp-config.php")
                )

        with doing("Restoring files"):
            remote = get_remote(config)
            restore_files(join(d, "wordpress"), remote)

        if config["git"]:
            with doing("Cloning Git repos"):
                for repo in config["git"]:
                    location = remote.child(repo["location"])
                    location.set_git_repo(repo["repo"], repo["version"])
                    doing.logger.info(
                        "Cloned %s@%s to %s", repo["repo"], repo["version"], location
                    )

        if config["owner"]:
            with doing("Changing files owner"):
                remote.chown(config["owner"])

        with doing("Reading WP config"):
            wp_config = get_wp_config(config)

        if config["mysql_root"]:
            mysql_root = config["mysql_root"]

            with doing("Ensuring that DB and user exist"):
                ensure_db_exists(wp_config, mysql_root, remote)

        with doing("Restoring DB"):
            db = create_from_source(wp_config, remote)
            restore_db(db, dump)

        if config["setup_queries"]:
            with doing("Running setup queries"):
                run_queries(db, config["setup_queries"])

        if config["outer_files"]:
            with doing("Creating outer files"):
                install_outer_files(config["outer_files"], remote)

        if config["post_install"]:
            with doing("Running post install scripts"):
                run_post_install(config["post_install"], remote)

        if config["dns"]:
            with doing("Configuring DNS"):
                configure_dns(config["dns"])


def __main__():
    return run_main(main, doing)


if __name__ == "__main__":
    __main__()



import json
from json import JSONDecodeError
from typing import Dict, List, Optional, Text

from luh3417.luhfs import Location, parse_location
from luh3417.luhsql import LuhSql, create_root_from_source
from luh3417.record_set import RecordSet, Zone, parse_domain
from luh3417.serialized_replace import ReplaceMap
from luh3417.snapshot import sync_files
from luh3417.utils import LuhError, escape


def read_config(file_path: Text):
    """
    Read configuration from JSON file (extracted from the snapshot)
    """

    try:
        with open(file_path, "r", encoding="utf-8") as f:
            return json.load(f)
    except json.JSONDecodeError:
        raise LuhError("Configuration file is not valid JSON")
    except OSError as e:
        raise LuhError(f"Error while opening file: {e}")


def get_remote(config: Dict) -> Location:
    """
    Reads the configuration to extract the address of the remote
    """

    try:
        return parse_location(config["args"]["source"])
    except KeyError:
        raise LuhError("Configuration is incomplete, missing args.source")


def get_wp_config(config: Dict) -> Dict:
    """
    Reads the configuration to extract the database configuration
    """

    try:
        return config["wp_config"]
    except KeyError:
        raise LuhError("Configuration is incomplete, missing wp_config")


def restore_files(wp_root: Text, remote: Location):
    """
    Restores the file from the local wp_root to the remote location
    """

    local = parse_location(wp_root)
    sync_files(local, remote, delete=True)


def restore_db(db: LuhSql, dump_path: Text):
    """
    Restores the specified file into DB, using the wp config and remote
    location to connect the DB.
    """

    try:
        with open(dump_path, "r", encoding="utf-8") as f:
            db.restore_dump(f)
    except OSError as e:
        raise LuhError(f"Could not read SQL dump: {e}")


def run_queries(db: LuhSql, queries: List[Text]):
    """
    Runs all the queries from the config
    """

    for query in queries:
        db.run_query(query)


def make_replace_map(replace_in_dump: List[Dict[Text, Text]]) -> ReplaceMap:
    """
    Transforms the config/patch syntax into an internal ReplaceMap
    """

    return [
        (x["search"].encode("utf-8"), x["replace"].encode("utf-8"))
        for x in replace_in_dump
    ]


def ensure_db_exists(wp_config, mysql_root, source: Location):
    """
    If a database is pre-existing, delete it. Then create a new one and create
    the appropriate user.
    """

    db = create_root_from_source(wp_config, mysql_root, source)

    name = escape(wp_config["db_name"], "`")
    user = escape(wp_config["db_user"], "`")
    host = escape(wp_config["db_host"], "`")
    password = escape(wp_config["db_password"], "'")

    db.run_query(f"drop database if exists {name};")
    db.run_query(f"create database {name};")
    db.run_query(f"create user if not exists {user}@{host} identified by {password};")
    db.run_query(f"alter user {user}@{host} identified by {password};")
    db.run_query(f"grant all on {name}.* to {user}@{host};")


def install_outer_files(outer_files: List[Dict], source: Location):
    """
    Given the list of outer files, sets the appropriate content to the
    appropriate location
    """

    for f in outer_files:
        location = source.child(f["name"])
        location.set_content(f["content"])


def run_post_install(post_install: List[Text], source: Location):
    """
    Running the post-install scripts
    """

    for script in post_install:
        source.run_script(script)


def configure_dns(dns: Dict):
    """
    Configures the DNS zones by loading all the zones into RecordSet and then
    calling RecordSet on all entries to configure them properly
    """

    zones = []

    for zone in dns["providers"]:
        zone["domain"] = parse_domain(zone["domain"])
        zones.append(Zone(**zone))

    rs = RecordSet(zones)

    for entry in dns["entries"]:
        if entry["type"] == "alias":
            func = rs.set_alias
        elif entry["type"] == "ips":
            func = rs.set_ips
        else:
            raise LuhError(f'Unknown entry type {entry["type"]}')

        func(**entry["params"])


def patch_config(
    config: Dict, patch_location: Optional[Location], allow_in_place: bool = False
) -> Dict:
    """
    Applies a configuration patch from the source patch file, which will
    alter the restoration process.

    Available options:

    - `wp_config` - WordPress configuration (database user and so on)
    - `source`
    - `owner` - Same syntax as chown owner, changes the ownership of restored
      files
    - `git` - A list of repositories to clone (cf below).
    - `setup_queries` - A list of SQL queries (as strings) that will be
      executed after restoring the DB
    - `php_define` - A dictionary of constant/value to be defined in wp-config
    - `replace_in_dump` - Replaces a list of values in the SQL dump
    - `mysql_root` - Method and options to become root of MySQL (see the
       README)
    - `outer_files` - Files to place on the host's filesystem
    - `post_install` - A list of Bash scripts to be executed when the install
      is done

    Example for the `git` value:

        "git": [
            {
                "location": "wp-content/themes/jupiter-child",
                "repo": "git@gitlab.com:your_company/jupiter_child.git",
                "version": "master"
            }
        ]

    Example for the `replace_in_dump` value:

        "replace_in_dump": [
            {
                "search": "https://old-domain.com",
                "replace": "https://new-domain.com"
            }
        ]

    Example for the `outer_files` value:

        "outer_files": [
            {
                "name": "robots.txt",
                "content": "User-agent: *\nDisallow: /\n"
            },
            {
                "name": "/etc/apache2/sites-available/my-host.conf",
                "content": "<VirtualHost> ..."
            }
        ]
    """

    base_config = {
        "owner": None,
        "git": [],
        "setup_queries": [],
        "php_define": {},
        "replace_in_dump": [],
        "mysql_root": None,
        "outer_files": [],
        "post_install": [],
        "dns": {},
    }

    for k, v in config.items():
        base_config[k] = v

    if patch_location:
        try:
            content = patch_location.get_content()
            patch = json.loads(content)
        except JSONDecodeError as e:
            raise LuhError(f"Could not decode patch file: {e}")
        else:
            if not allow_in_place:
                try:
                    assert patch["args"]["source"]
                except (AssertionError, KeyError):
                    raise LuhError(
                        "The patch did not override the source location "
                        "and the --allow-in-place flag is not set"
                    )

            for k, v in patch.items():
                base_config[k] = v
    elif not allow_in_place:
        raise LuhError(
            "If you do not set the --allow-in-place flag you must provide a "
            "patch which overrides the source location"
        )

    return base_config



from dataclasses import dataclass, replace
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import Dict, List, Optional, Sequence, Text, Union

from libcloud.dns.base import DNSDriver, Record
from libcloud.dns.base import Zone as CloudZone
from libcloud.dns.providers import get_driver
from libcloud.dns.types import RecordType

IpAddress = Union[IPv4Address, IPv6Address, Text]


class RecordSetError(Exception):
    pass


def parse_domain(domain: Union[Text, "Domain"]) -> "Domain":
    """
    Transforms a domain as a text string into a Domain instance. Calling with
    a domain will return a copy of the said domain
    """

    if isinstance(domain, Domain):
        return replace(domain)
    else:
        parts = tuple(p.lower() for p in reversed(domain.split(".")))

        if any(not p for p in parts):
            raise RecordSetError(f"Domain {domain} does not appear to be valid")

        return Domain(domain, tuple(p.lower() for p in reversed(domain.split("."))))


@dataclass
class Domain:
    """
    Lightly-parsed domain name, mostly here to do some basic operations (see
    below)
    """

    domain: Text
    parts: Sequence[Text]

    def __str__(self):
        """
        Transforms back the domain into a string
        """

        return self.domain

    def __eq__(self, other: Union[Text, "Domain"]):
        """
        Tests the equality domain-wise (aka case-insensitive)
        """

        other = parse_domain(other)
        return self.parts == other.parts

    def contains(self, other: Union[Text, "Domain"]) -> bool:
        """
        Tests of the other domain is a sub-domain or equal to the current
        domain
        """

        other = parse_domain(other)

        try:
            for i, part in enumerate(self.parts):
                if other.parts[i] != part:
                    return False
        except IndexError:
            return False
        else:
            return True

    def truncate(self, other: Union[Text, "Domain"]) -> "Domain":
        """
        Returns the other domain without the current domain. By example:

        >>> d = parse_domain('foo.com')
        >>> assert d.truncate('bar.foo.com') == 'bar'
        """

        other = parse_domain(other)

        if not self.contains(other):
            raise RecordSetError(
                f"Cannot truncate {other} with {self} because not contained"
            )

        i = 0

        try:
            for i, part in enumerate(self.parts):
                if other.parts[i] != part:
                    break
        except IndexError:
            pass

        out = other.parts[i + 1 :]

        return Domain(domain=".".join(reversed(out)), parts=tuple(out))


def list_or_iterate(obj, word, *args, **kwargs):
    """
    Tries both the iterate and the list methods from libcloud (wtf aren't both
    implemented?)
    """

    try:
        return getattr(obj, f"iterate_{word}")(*args, **kwargs)
    except NotImplementedError:
        return getattr(obj, f"list_{word}")(*args, **kwargs)


@dataclass
class Zone:
    """
    A DNS zone, represented by its domain (like my.org), a provider (like
    digitalocean) and credentials which is a kwarg dict which will be passed
    to the constructor of the provider.
    """

    domain: Domain
    provider: Text
    credentials: Dict[Text, Text]

    def get_cloud_zone(self, driver: DNSDriver) -> Optional[CloudZone]:
        """
        Go get the equivalent libcloud zone from the driver
        """

        for zone in list_or_iterate(driver, "zones"):
            if self.domain == zone.domain:
                return zone


class RecordSet:
    """
    A class to manage records by set instead of creating them alone
    """

    def __init__(self, zones: Sequence[Zone]):
        self.zones = zones

    def _get_zone_info(self, domain):
        """
        Computes and checks some zone info
        """

        zone = self.get_zone(domain)

        if not zone:
            raise RecordSetError(f"Cannot find a zone for {domain}")

        driver = self.get_driver(zone)
        cloud_zone = zone.get_cloud_zone(driver)

        if not cloud_zone:
            raise RecordSetError(f"Cannot find the zone for {domain} in the provider")

        sub_domain = zone.domain.truncate(domain)

        return cloud_zone, driver, sub_domain

    def get_zone(self, domain: Union[Domain, Text]) -> Optional[Zone]:
        """
        Given a domain, returns the first matching zone (aka zone which
        contains the provided domain)
        """

        domain = parse_domain(domain)

        for zone in self.zones:
            if zone.domain.contains(domain):
                return zone

    def get_driver(self, zone: Zone) -> DNSDriver:
        """
        Given a zone, starts the appropriate driver
        """

        cls = get_driver(zone.provider)
        return cls(**zone.credentials)

    def set_ips(self, domain: Union[Domain, Text], ips: Sequence[IpAddress]):
        """
        Given a domain and a set of IP addresses, defines the correct A and
        AAAA records while also deleting the extraneous one (and CNAME too)
        """

        cloud_zone, driver, sub_domain = self._get_zone_info(domain)

        expected_ips = set(ip_address(ip) for ip in ips)
        found_ips = set()
        to_delete: List[Record] = []

        for record in list_or_iterate(driver, "records", cloud_zone):  # type: Record
            if sub_domain == record.name:
                if record.type in {RecordType.A, RecordType.AAAA}:
                    ip = ip_address(record.data)
                    found_ips.add(ip)

                    if ip not in expected_ips:
                        to_delete.append(record)
                elif record.type == RecordType.CNAME:
                    to_delete.append(record)

        for record in to_delete:
            record.delete()

        for ip in expected_ips - found_ips:
            driver.create_record(
                name=f"{sub_domain}",
                zone=cloud_zone,
                type=RecordType.A if ip.version == 4 else RecordType.AAAA,
                data=f"{ip}",
            )

    def set_alias(self, domain: Union[Domain, Text], target: Union[Text, Domain]):
        """
        Sets the domain to be an alias. It will first remove any AAAA or A
        record attached to this domain.
        """

        cloud_zone, driver, sub_domain = self._get_zone_info(domain)
        target = parse_domain(target)
        to_delete: List[Record] = []
        to_check: Optional[Record] = None

        for record in list_or_iterate(driver, "records", cloud_zone):  # type: Record
            if sub_domain == record.name:
                if record.type in {RecordType.A, RecordType.AAAA}:
                    to_delete.append(record)
                elif record.type == RecordType.CNAME:
                    to_check = record

        for record in to_delete:
            record.delete()

        if to_check:
            if target != to_check.data:
                to_check.update(data=f"{target}.")
        else:
            driver.create_record(
                name=f"{sub_domain}",
                data=f"{target}.",
                type=RecordType.CNAME,
                zone=cloud_zone,
            )



import json
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from os.path import exists
from tempfile import NamedTemporaryFile
from typing import Optional, Sequence, Text

from luh3417.luhfs import parse_location
from luh3417.luhphp import parse_wp_config
from luh3417.restore.__main__ import main as restore
from luh3417.snapshot.__main__ import main as snapshot
from luh3417.transfer import UnknownEnvironment, apply_wp_config
from luh3417.utils import import_file, make_doer, run_main, setup_logging

doing = make_doer("luh3417.transfer")


def generator_validator(file_path: Text):
    """
    Validates that the generator module exists and has all the required
    methods
    """

    if not exists(file_path):
        raise ArgumentTypeError(f"File {file_path} could not be found")

    try:
        module = import_file("generator", file_path)
    except SyntaxError:
        doing.logger.exception("Syntax error in generator")
        raise ArgumentTypeError(f"File {file_path} has a syntax error")
    except ImportError:
        raise ArgumentTypeError(f"File {file_path} cannot be imported")
    except Exception:
        doing.logger.exception("Unknown error while importing generator")
        raise ArgumentTypeError(f"Unknown error")

    if not hasattr(module, "get_source") or not callable(module.get_source):
        raise ArgumentTypeError(
            f"Generator does not expose a get_source(environment_name) method"
        )

    if not hasattr(module, "allow_transfer") or not callable(module.allow_transfer):
        raise ArgumentTypeError(
            f"Generator does not expose a allow_transfer(origin, target) method"
        )

    if not hasattr(module, "get_backup_dir") or not callable(module.get_backup_dir):
        raise ArgumentTypeError(
            f"Generator does not expose a get_backup_dir(environment) method"
        )

    if not hasattr(module, "get_patch") or not callable(module.get_backup_dir):
        raise ArgumentTypeError(
            f"Generator does not expose a get_patch(origin, target) method"
        )

    if not hasattr(module, "get_wp_config") or not callable(module.get_backup_dir):
        raise ArgumentTypeError(
            f"Generator does not expose a get_wp_config(environment) method"
        )

    return module


def parse_args(args: Optional[Sequence[str]] = None) -> Namespace:
    """
    Parses args and validates the consistency of origin/target using the
    generator
    """

    parser = ArgumentParser(
        prog="python -m luh3417.transfer",
        description="Transfers a WordPress to one location to the other",
    )

    parser.add_argument(
        "-g",
        "--settings-generator",
        help="A Python script that handles the transitions",
        type=generator_validator,
        required=True,
    )

    parser.add_argument("origin", help="Origin environment")
    parser.add_argument("target", help="Target environment")

    parsed = parser.parse_args(args)

    for env in ["origin", "target"]:
        env_name = getattr(parsed, env)

        try:
            parsed.settings_generator.get_source(env_name)
        except UnknownEnvironment as e:
            parser.error(
                f'Environment "{env_name}" not recognized by generator: {e.message}'
            )

    if not parsed.settings_generator.allow_transfer(parsed.origin, parsed.target):
        parser.error(
            f"Generator does not allow transfer from {parsed.origin} to {parsed.target}"
        )

    return parsed


def main(args: Optional[Sequence[str]] = None):
    """
    Executes things in order
    """

    setup_logging()
    args = parse_args(args)

    gen = args.settings_generator

    origin_source = parse_location(gen.get_source(args.origin))
    origin_backup_dir = gen.get_backup_dir(args.origin)

    with doing(f"Backing up {args.origin} to {origin_backup_dir}"):
        origin_archive = snapshot([f"{origin_source}", origin_backup_dir])

    target_backup_dir = gen.get_backup_dir(args.target)
    target_source = parse_location(gen.get_source(args.target))

    with doing(f"Checking if {args.target} ({target_source}) already exists"):
        target_exists = target_source.exists()

    if target_exists:
        with doing(f"Backing up {args.target} to {target_backup_dir}"):
            snapshot([f"{target_source}", target_backup_dir])

    if target_exists:
        with doing(f"Reading wp_config from {args.target}"):
            wp_config = parse_wp_config(target_source)
    else:
        with doing(f"Generating wp_config for {args.target}"):
            wp_config = gen.get_wp_config(args.target)

    patch = apply_wp_config(
        gen.get_patch(args.origin, args.target), wp_config, target_source
    )

    with NamedTemporaryFile(mode="w", encoding="utf-8") as pf:
        json.dump(patch, pf)
        pf.flush()

        with doing(f"Overriding {args.target} with {args.origin}"):
            restore(["-p", pf.name, f"{origin_archive}"])

    if hasattr(gen, "post_exec"):
        with doing("Running post-exec hook"):
            gen.post_exec(args.origin, args.target)


def __main__():
    return run_main(main, doing)


if __name__ == "__main__":
    __main__()



import json
import re
import subprocess
from itertools import chain
from typing import Any, Dict, Text

from luh3417.luhfs import Location
from luh3417.utils import LuhError

PHP_STR = (
    r"\"(?:[^\"\\]|\\(?:[nrtvef\\$\"]|[0-7]{1,3}|x[0-9A-Fa-f]{1,2}|u[0-9A-Fa-f]+))*\""
    r"|'(?:[^'\\]|\\')*'"
)


def parse_wp_config(location: "Location", config_file_name: Text = "wp-config.php"):
    """
    Parses the WordPress configuration to get the DB configuration
    """

    config_location = location.child(config_file_name)
    config = config_location.get_content()
    const = extract_php_constants(config)

    try:
        return {
            "db_host": const["DB_HOST"],
            "db_user": const["DB_USER"],
            "db_password": const["DB_PASSWORD"],
            "db_name": const["DB_NAME"],
        }
    except KeyError as e:
        LuhError(f"Missing config value: {e}")


def run_php(code: str):
    """
    Runs PHP code and returns its output or None if there was an error
    """

    cp = subprocess.run(
        ["php"],
        input=code,
        stderr=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        encoding="utf-8",
    )

    if cp.returncode:
        return None

    return cp.stdout


def extract_php_constants(file: str):
    """
    Parses a PHP file to extract all the declared constants
    """

    define = re.compile(r"^\s*define\(")
    lines = ["<?php"]

    for line in file.splitlines(False):
        if define.match(line):
            lines.append(line)

    lines.extend(
        [
            "$const = get_defined_constants(true);",
            '$user = $const["user"];',
            "echo json_encode($user);",
        ]
    )

    try:
        data = run_php("\n".join(lines))
        return json.loads(data)
    except (ValueError, TypeError):
        raise LuhError("Configuration file has syntax errors")


def parse_php_string(string: Text) -> Text:
    """
    Parses a PHP string literal (including quotes) and returns the equivalent
    Python string (as a string, not a literal)
    """

    parser = f"<?php echo json_encode({string});"

    try:
        data = run_php(parser)
        return json.loads(data)
    except (ValueError, TypeError):
        raise LuhError("Invalid PHP string")


def encode_php_string(string: Text) -> Text:
    """
    Encodes a string into something that can be given to PHP as a string
    literal
    """

    quoted = string.replace("'", "\\'")
    return f"'{quoted}'"


def encode_php_value(val):
    """
    Transforms a JSON-encodable value into a PHP literal
    """

    data = encode_php_string(json.dumps(val))
    parser = f"<?php var_export(json_decode({data}, true));"

    try:
        data = run_php(parser)
        assert data is not None
        return data
    except AssertionError:
        raise LuhError(f"Could not encode to PHP value: {repr(val)}")


def set_wp_config_values(values: Dict[Text, Any], file_path: Text):
    """
    Makes sure that the key/values set in values are set to this value in
    wp-config.php.

    - Existing define() are modified to get the new value
    - Missing define() are added at the top of the file

    Values must be JSON-encodable
    """

    try:
        e = re.compile(r"^\s*define\(\s*(" + PHP_STR + ")")
        out = []
        patched = set()

        with open(file_path, "r", encoding="utf-8") as f:
            for line in (x.rstrip("\r\n") for x in f.readlines()):
                m = e.match(line)

                if e.match(line):
                    key = parse_php_string(m.group(1))
                else:
                    key = None

                if key in values:
                    out.append(
                        f"define("
                        f"{encode_php_string(key)}, "
                        f"{encode_php_value(values[key])}"
                        f");"
                    )
                    patched.add(key)
                else:
                    out.append(line)

        missing = set(values.keys()) - patched
        extra = []

        if missing:
            extra = ["/** Extra values created by LUH3417 */"]

            for k in missing:
                v = values[k]
                extra.append(
                    f"define("
                    f"{encode_php_string(k)}, "
                    f"{encode_php_value(v)}"
                    f");"
                )

            extra.append("")

        with open(file_path, "w", encoding="utf-8") as f:
            f.write("\n".join(chain(out[0:1], extra, out[1:])) + "\n")
    except OSError as e:
        raise LuhError(f"Could not open config file at {file_path}: {e}")



from contextlib import contextmanager
from importlib.util import module_from_spec, spec_from_file_location
from logging import DEBUG, getLogger
from random import SystemRandom
from typing import Text

import coloredlogs

random = SystemRandom()


class LuhError(Exception):
    """
    Managed errors that are emitted by the code in order to be displayed nicely
    (aka without stack trace) to the user. Basically if you know that something
    can go wrong then write a nice explicative message and throw it in a
    LuhError when it happens.
    """

    def __init__(self, message):
        self.message = message


def setup_logging():
    """
    Setups the log formatting
    """

    coloredlogs.install(
        level=DEBUG, fmt="%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s"
    )


def make_doer(name):
    """
    Generates the logger and the doing() context manager for a given name.

    The doing() context manager will display a message in the logs and handle
    exceptions occurring during execution, displaying them in the logs as well.

    If an exception occurs, the program is exited.
    """

    logger = getLogger(name)

    @contextmanager
    def doing(message):
        logger.info(message)

        # noinspection PyBroadException
        try:
            yield
        except LuhError as e:
            logger.error(e.message)
            exit(1)
        except Exception:
            logger.exception("Unknown error")
            exit(2)

    doing.logger = logger

    return doing


def escape(string, char):
    """
    Quotes a string with the given char. By example:

    >>> assert escape("O'Neil", "'") == "'O\\'Neil'"
    """
    return char + string.replace(char, f"\\{char}") + char


def import_file(name: Text, file_path: Text):
    """
    Imports a Python file as a module named luh3417.{name}
    """

    spec = spec_from_file_location(f"luh3417.{name}", file_path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def random_password(
    n=50, chars="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890/>#"
):
    """
    Generates a secure random password
    """

    return "".join(random.choice(chars) for _ in range(0, n))


def run_main(main, doing):
    """
    Runs a main() function while taking care to catch the appropriate
    exceptions and do the cleaning up
    """

    # noinspection PyBroadException
    try:
        main()
    except KeyboardInterrupt:
        doing.logger.info("Quitting due to user signal")
    except SystemExit:
        pass
    except BaseException:
        doing.logger.exception("Unknown error")
    finally:
        from luh3417.luhssh import SshManager

        SshManager.shutdown()



import subprocess
from typing import Sequence, Text

from luh3417.luhfs import LocalLocation, Location, SshLocation
from luh3417.luhssh import SshManager
from luh3417.utils import LuhError


def sync_files(remote: Location, local: Location, delete: bool = False):
    """
    Use rsync to copy files from a location to another
    """

    local.ensure_exists_as_dir()

    args = [
        "rsync",
        "-rz",
        "--exclude=.git",
        "--exclude=.idea",
        "--exclude=*.swp",
        "--exclude=*.un~",
    ]

    if delete:
        args.append("--delete")

    args += [remote.rsync_path(True), local.rsync_path(True)]

    cp = subprocess.run(args, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)

    if cp.returncode:
        raise LuhError(f"Error while copying files: {cp.stderr}")


def _build_args(location: Location, args: Sequence[Text]) -> Sequence[Text]:
    """
    Builds args to use either with SSH either straight
    """

    if isinstance(location, LocalLocation):
        return args
    elif isinstance(location, SshLocation):
        return SshManager.instance(location.user, location.host).get_args(args)


def copy_files(remote: Location, local: Location):
    """
    Copies files from the remote location to the local locations. Files are
    serialized and pipelined through tar, maybe locally, maybe through SSH
    depending on the locations.
    """

    remote_args = _build_args(
        remote, ["tar", "--warning=no-file-changed", "-C", remote.path, "-c", "."]
    )
    local_args_1 = _build_args(local, ["mkdir", "-p", local.path])
    local_args_2 = _build_args(local, ["tar", "-C", local.path, "-x"])

    cp = subprocess.run(local_args_1, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)

    if cp.returncode:
        raise LuhError(f'Error while creating target dir "{local}": {cp.stderr}')

    remote_p = subprocess.Popen(
        remote_args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    local_p = subprocess.Popen(
        local_args_2,
        stdin=remote_p.stdout,
        stderr=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
    )

    remote_p.wait()
    local_p.wait()

    if remote_p.returncode:
        err = remote_p.stderr.read(1000)

        if err == b"":
            return

        raise LuhError(f'Error while reading files from "{remote}": {err}')

    if local_p.returncode:
        raise LuhError(f'Error writing files to "{local}": {local_p.stderr.read(1000)}')



import json
from argparse import ArgumentParser, Namespace
from datetime import datetime
from os.path import join
from tempfile import TemporaryDirectory
from typing import Dict, Optional, Sequence, Text

from luh3417.luhfs import Location, parse_location
from luh3417.luhphp import parse_wp_config
from luh3417.luhsql import create_from_source
from luh3417.snapshot import copy_files
from luh3417.utils import make_doer, run_main, setup_logging

doing = make_doer("luh3417.snapshot")


def parse_args(args: Optional[Sequence[str]] = None) -> Namespace:
    """
    Parse arguments fro the snapshot
    """

    parser = ArgumentParser(
        description=(
            "Takes a snapshot of a WordPress website remotely and "
            "stores it to either a local or a remote location. This "
            "requires rsync, php-cli and mysqldump"
        )
    )

    parser.add_argument(
        "source",
        help=(
            "Source directory for the WordPress installation dir. Syntax: "
            "`/var/www` or `user@host:/var/www`"
        ),
        type=parse_location,
    )
    parser.add_argument(
        "backup_dir", help="Directory to store the snapshot", type=parse_location
    )
    parser.add_argument(
        "-n",
        "--snapshot-base-name",
        help="Base name for the snapshot file. Defaults to DB name.",
    )
    parser.add_argument(
        "-t",
        "--file-name-template",
        help="Template for snapshot file name. Defaults to: `{base}_{time}.tar.gz`",
        default="{base}_{time}.tar.gz",
    )

    return parser.parse_args(args)


def make_dump_file_name(args: Namespace, wp_config: Dict, now: datetime) -> Location:
    """
    Generates the location name where to dump the file
    """

    if not args.snapshot_base_name:
        base_name = wp_config["db_name"]
    else:
        base_name = args.snapshot_base_name

    name = args.file_name_template.format(base=base_name, time=now.isoformat() + "Z")

    return args.backup_dir.child(name)


def dump_settings(args: Namespace, wp_config: Dict, now: datetime, file_path: Text):
    """
    Given the settings and various environmental data, dump them in a JSON file
    which will be embedded in the archive and can be used to guess things when
    restoring the snapshot.
    """

    args = dict(vars(args), source=f"{args.source}", backup_dir=f"{args.backup_dir}")

    content = {"args": args, "wp_config": wp_config, "time": now.isoformat() + "Z"}

    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(content, f, indent=4)


def main(args: Optional[Sequence[str]] = None):
    """
    Executes things in order
    """

    setup_logging()
    args = parse_args(args)
    now = datetime.utcnow()

    with doing("Parsing remote configuration"):
        wp_config = parse_wp_config(args.source)

    with TemporaryDirectory() as d:
        work_location = parse_location(d)

        with doing("Saving settings"):
            dump_settings(args, wp_config, now, join(d, "settings.json"))

        with doing("Copying database"):
            db = create_from_source(wp_config, args.source)
            db.dump_to_file(join(d, "dump.sql"))

        with doing("Copying files"):
            copy_files(args.source, work_location.child("wordpress"))

        with doing("Writing archive"):
            args.backup_dir.ensure_exists_as_dir()
            archive_location = make_dump_file_name(args, wp_config, now)

            archive_location.archive_local_dir(d)
            doing.logger.info("Wrote archive %s", archive_location)

    return archive_location


def __main__():
    return run_main(main, doing)


if __name__ == "__main__":
    __main__()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['luh3417',
 'luh3417.replace',
 'luh3417.restore',
 'luh3417.snapshot',
 'luh3417.transfer']

package_data = \
{'': ['*']}

install_requires = \
['apache-libcloud>=2.0,<3.0', 'coloredlogs>=9.0']

entry_points = \
{'console_scripts': ['luh3417_replace = luh3417.replace.__main__:__main__',
                     'luh3417_restore = luh3417.restore.__main__:__main__',
                     'luh3417_snapshot = luh3417.snapshot.__main__:__main__',
                     'luh3417_transfer = luh3417.transfer.__main__:__main__']}

setup_kwargs = {
    'name': 'luh3417',
    'version': '0.1.5',
    'description': 'A WordPress backup/restore/workflow tool',
    'long_description': 'LUH3417\n=======\n\nThis is a tool to help you implement a WordPress development workflow. It has 3\nmain features:\n\n- **Snapshot** &mdash; Take snapshots of a running WordPress instance\n- **Restore** &mdash; Restore those snapshots in-place or to a different\n  location\n- **Transfer** &mdash; Transfer one instance over another using automated\n  backup, validation and configuration rules\n\nEverything can happen seamlessly in *local* or *through SSH*, allowing you to\nwork easily on remote servers from your local machine and to transfer instances\nfrom one server to another.\n\nThanks to this, putting your code to production is as simple as:\n\n```bash\npython -m luh3417.transfer -g my_project.py local production\n```\n\nWhile the `snapshot` and `restore` operations can be used individually, it is\nnot recommended to use them as the main tools. Indeed, `restore` can easily\noverride an instance without any previous backup. For this reason, it is better\nto use `transfer` whenever possible. It will ensure your safety within the\nworkflow that you defined.\n\n## Installation\n\n```\npip install luh3417\n```\n\n## Usage\n\nLUH3417 is made to use with Python\'s `-m` option. This way, if you want to\ninvoke the `snapshot` feature, the base command will be\n`python -m luh3417.snapshot`.\n\nIf you prefer, there is also equivalent commands installed in the `bin`\ndirectory, namely `luh3417_snapshot`, `luh3417_restore` and `luh3417_transfer`.\n\nAll the locations can be in two formats:\n\n- `SSH` &mdash; `user@server:/location/on/server`\n- `Local` &mdash; `/location/on/current/machine`\n\nThis allows you to transfer data between remote servers and local machine quite\nseamlessly.\n\n> **NOTE** &mdash; You need to use an SSH agent in order for all the features\n> to work. No password prompt will show up. Usually it\'s as simple as to type\n> `ssh-add` in your terminal once during your session.\n\n### `snapshot`\n\nCreates a snapshot of a running WordPress instance. A snapshot is an archive\ncontaining:\n\n- All PHP/theme/media/etc files\n- A DB dump\n- Meta information about how the snapshot was taken\n\nUsage syntax:\n\n```\npython -m luh3417.snapshot [-h] [-n SNAPSHOT_BASE_NAME] [-t FILE_NAME_TEMPLATE] source backup_dir\n```\n\nExample:\n\n```\npython -m luh3417.snapshot root@prod-server.com:/var/www/html root@backup-server.com:/var/backups/wp\n```\n\nAdditional options:\n\n- `-n`/`--snapshot-base-name` &mdash; Base name for your snapshot file. See\n  the `--file-name-template` option to see how this name is used. The default\n  name is the database\'s name.\n- `-t`/`--file-name-template` &mdash; This template will be used to generate\n  the snapshot file name. By default it is `{base}_{time}.tar.gz` but you can\n  put whatever you want. `{base}` and `{time}` will be replaced respectively\n  by the base name (see `--snapshot-base-name`) and the ISO 8601 UTC date.\n  Independently of the name, the file will be placed in the `backup_dir`.\n\n### `restore`\n\nRestores a snapshot either in-place to its original location using the embedded\nmeta-data or to another location using a patch on the meta-data.\n\nIn addition to just restoring the files and database, the patch can trigger\nchanges in `wp-settings.php`, replace values in the database and much more.\n\n**`restore` will essentially override an instance with the content of a\nbackup, so make sure to use it wisely in order not to loose data. Also, see\n`transfer`**.\n\nUsage:\n\n```\npython -m luh3417.restore [-p PATCH] [-a ALLOW_IN_PLACE] snapshot\n```\n\nOptions:\n\n- `-p`/`--patch` &mdash; Location to the patch file (see below)\n- `-a`/`--allow-in-place` &mdash; Allows restoring the backup onto its original\n  location. This flag is required because otherwise it would be way too easy\n  to override\n\n#### Restore in-place\n\nIf you want to restore a backup to its original location, you just need to\nknow the file\'s location and pass the `-a` flag.\n\n```\npython -m luh3417.restore -a root@backup-server.com:/path/to/snapshot.tar.gz\n```\n\n> **NOTE** &mdash; If the snapshot was made locally, it will always be restored\n> locally because there is no way for LUH3417 to know the originating server so\n> it assumes that the snapshot file was not transferred to another machine.\n\n#### Restore to another location\n\nIn order to restore to another location, you need to use a patch file\n\n```\npython -m luh3417.restore -p patch.json root@backup-server.com:/path/to/snapshot.tar.gz\n```\n\nHere is an example of patch file:\n\n```json\n{\n    "args": {\n        "source": "root@new-server.com:/var/www/html"\n    },\n    "owner": "www-data:"\n}\n```\n\nSee below for detailed documentation of patch content\n\n##### `args.source`\n\nSet this one to define where to restore the archive.\n\n```json\n{\n    "args": {\n        "source": "root@new-server.com:/var/www/html"\n    }\n}\n```\n\n##### `wp_config`\n\nDatabase configuration from the WordPress\n\n```json\n{\n    "wp_config": {\n        "db_host": "localhost",\n        "db_name": "xxx",\n        "db_user": "xxx",\n        "db_password": "xxx"\n    }\n}\n```\n\n> **NOTE** &mdash; You need to make sure you match those values in `php_define`\n> unless you\'re using `transfer` which sets them automatically\n\n##### `owner`\n\nThis changes the owner of the files to another one. This only works if:\n\n- When restoring locally, you run as `root`\n- When restoring remotely, you login in as `root`\n\n```json\n{\n    "owner": "www-data:"\n}\n```\n\n##### `git`\n\nReplaces some directories with a Git repository at a given version\n\n```json\n{\n    "git": [\n        {\n            "location": "wp-content/themes/jupiter-child",\n            "repo": "git@gitlab.com:your_company/jupiter_child.git",\n            "version": "master"\n        }\n    ]\n}\n```\n\n> **NOTE** &mdash; `.git` directories are excluded from snapshots, so unless\n> you specify this option there will be no git-enabled directories in the\n> restored files. On the other hand, git repositories will be created at\n> specified version, so it might not make sense to specify this option when\n> restoring a backup in-place.\n\n##### `setup_queries`\n\nA list of SQL queries to be run after the DB was restored\n\n```json\n{\n    "setup_queries": [\n        "delete from wp_options where option_name = \'gtm4wp-options\';"\n    ]\n}\n```\n\n##### `php_define`\n\nValues to be changed or added in `wp-config.php`. Any JSON-serializable value\ncan be used.\n\n```json\n{\n    "php_define": {\n        "WP_CACHE": false,\n        "WP_SENTRY_ENV": "new-env"\n    }\n}\n```\n\n##### `replace_in_dump`\n\nA list of strings with their replacement to be changed in the dump before\nrestoring it. This is mainly used to change the domain name of the instance.\nAs WordPress serializes its settings, a simple replace is not possible. This\nwill use a holistic heuristic which will try to keep PHP-serialized values\ncorrect even if quoted in a MySQL string.\n\n> **NOTE** &mdash; PHP-serialized values are prefixed by their length, this is\n> why a simple replace cannot be effective: if the length changes then the\n> whole value gets corrupted.\n\n```json\n{\n    "replace_in_dump": [\n        {\n            "search": "https://old-domain.com",\n            "replace": "https://new-domain.com"\n        }\n    ]\n}\n```\n\n##### `mysql_root`\n\nIn order to create the database and set the user password, the script needs\na root access to MySQL. Today, the only supported method is `socket`, because\nit is password-less. However it only works when the server is local and\nproperly configured (it\'s the default behavior in Debian-based distros).\n\n```json\n{\n    "mysql_root": {\n        "method": "socket",\n        "options": {\n            "sudo_user": "root",\n            "mysql_user": "root"\n        }\n    }\n}\n```\n\nAbout the options:\n\n- `sudo_user` &mdash; don\'t set it if you don\'t need to sudo to use the socket,\n  set it to `root` or whichever user is right otherwise.\n- `mysql_user` &mdash; name of the MySQL user to use\n\n##### `outer_files`\n\nCreates files on the server\'s file system. If the file name is relative then\nthe file is created relatively to the WordPress\'s root, otherwise it is created\nat the specified absolute location.\n\n```json\n{\n    "outer_files": [\n        {\n            "name": "robots.txt",\n            "content": "User-agent: *\\nDisallow: /\\n"\n        },\n        {\n            "name": "/etc/apache2/sites-available/my-host.conf",\n            "content": "<VirtualHost> ..."\n        }\n    ]\n}\n```\n\n> **NOTE** &mdash; There is not (yet) any form of privilege escalation to\n> create those files, so the local/remote user must have the rights to create\n> those files.\n\n##### `post_install`\n\nThose are shell scripts which run on the host server after the install is\ncomplete. Typically, you can enable your virtual host and reload Apache.\n\n```json\n{\n    "post_install": [\n        "a2ensite my-website.com",\n        "systemctl reload apache2"\n    ]\n}\n```\n\n##### `dns`\n\nYou might want to use your DNS provider\'s API in order to configure the domain\nthat is going to target your website. LUH3417 integrates with\n[libcloud](https://libcloud.readthedocs.io/en/latest/index.html) in order to\nprovide an abstraction over the most popular cloud providers.\n\nHere is an example entry:\n\n```json\n{\n    "dns": {\n        "providers": [\n            {\n                "domain": "my-corp.net",\n                "provider": "digitalocean",\n                "credentials": {\n                    "key": "xxxxxx",\n                }\n            }\n        ],\n        "entries": [\n            {\n                "type": "alias",\n                "params": {\n                    "domain": "my-wp.my-corp.net",\n                    "target": "load-balancer.my-corp.net"\n                }\n            },\n            {\n                "type": "ips",\n                "params": {\n                    "domain": "dns.my-corp.net",\n                    "ips": [\n                        "2606:4700:4700::1111",\n                        "2606:4700:4700::1001",\n                        "1.1.1.1",\n                        "1.0.0.1"\n                    ]\n                }\n            }\n        ]\n    }\n}\n```\n\nLet\'s break this down\n\n###### `providers`\n\nThat\'s a list of the providers, associated to a domain name. The different keys\nare used like this:\n\n- `domain` &mdash; root domain name managed by this provider\n- `provider` &mdash; domain name provider (you can get the list\n  [here](https://github.com/apache/libcloud/blob/trunk/libcloud/dns/types.py#L32),\n  use the lower-case string value)\n- `credentials` &mdash; kwargs to be passed to the constructor of the provider\n\n###### `entries`\n\nEntries are either a single CNAME either a set of A/AAAA records for a same\ndomain name. LUH3417 will make sure that all records for this (sub-)domain\nmatch your specification and **will delete other records for that sub-domain**.\n\nSuppose the following situation:\n\n- `foo.my.org` resolves to `A 1.2.3.4`\n- But you want it to be a CNAME of `bar.my.org`\n- The `A 1.2.3.4` entry will be deleted and a `CNAME bar.my.org` will be\n  created\n\nNow, let\'s dig into the options\n\n**`"type" = "alias"`**\n\nThat\'s when you want to create a CNAME.\n\n```json\n{\n    "type": "alias",\n    "params": {\n        "domain": "my-wp.my-corp.net",\n        "target": "load-balancer.my-corp.net"\n    }\n}\n```\n\nThe two params are:\n\n- `domain` &mdash; target (sub-)domain\n- `target` &mdash; target of the CNAME (aka the value of the record)\n\n**`"type" = "ips"`**\n\nThis will set your (sub-)domain to point on a set if IP addresses, preferably\nv6 but legacy systems like v4 are still supported.\n\n```json\n{\n    "type": "ips",\n    "params": {\n        "domain": "dns.my-corp.net",\n        "ips": [\n            "2606:4700:4700::1111",\n            "2606:4700:4700::1001",\n            "1.1.1.1",\n            "1.0.0.1"\n        ]\n    }\n}\n```\n\n- `domain` &mdash; is the target (sub-)domain\n- `ips` &mdash; is a list of IP address that will be set to AAAA and A records\n\n### `transfer`\n\nThe main goal of this package is to allow the setup of a custom workflow that\nallows easy copy of WordPress instances from an environment from the other.\n\nThe basic idea is the following:\n\n- You can specify an origin and target environment names\n- There is a *settings generator* Python file which will generate all the\n  settings and patches appropriate for this transfer.\n\nIt\'s **your responsibility** to write an settings generator, however there is\nan a documented example attached in this repository.\n\nUsage:\n\n```\npython -m luh3417.transfer [-h] -g SETTINGS_GENERATOR origin target\n```\n\nExample:\n\n```\npython -m luh3417.transfer -g example/generator.py develop local\n```\n\nTo see the content of the generator file, please refer to the\n[example/generator.py](example/generator.py) file and especially the\n`allow_transfer()` method\'s documentation which will explain the spirit of\nthe file.\n\n### `replace`\n\nSeeks and replaces serialized values. Values could be in quoted MySQL literals\nor serialized PHP values, they will be replaced and the containing string will\nbe re-serialized.\n\nThat\'s a complicated problem due to PHP serialized format which prefixes the\nstring\'s length to the string. If you want to replace a value you need to\nreplace its length.\n\nBy example:\n\n```php\necho serialize("foo");\n// s:3:"foo";\n\necho serialize("long foo");\n// s:8:"long foo";\n//   ^-- see here how the prefix changed from 3 to 8\n```\n\nThis command will do the replacing in a file and output to another file. Be\ncareful, it will not warn you of overwrites.\n\nUsage:\n\n```\npython -m luh3417.replace [-h] -i INPUT -o OUTPUT [-b BEFORE [BEFORE ...]] [-a AFTER [AFTER ...]] [-c CHARSET]\n```\n\nExample:\n\n```\npython -m luh3417.replace \\\n    -i dump.sql \\\n    -o dump_replaced.sql \\\n    -b old_domain.com \\\n    -a new_domain.com\n```\n\n## FAQ\n\n> Why the name `LUH3417`?\n\nIt\'s a character from THX1138. The author is not particularly fan of this\nmovie, however it expresses quite well the feeling of working with WordPress\nand especially setting up a professional workflow.\n\n> Why using Python to code it?\n\nIt felt to the author that this language was more appropriate for this task\nthan PHP.\n\n> Do I need to write Python to use the transfer feature?\n\nYes, fortunately it\'s pretty easy. The author started with\n[Dive Into Python](https://www.diveinto.org/python3/).\n\n> Why can\'t the transfer feature have a configuration file instead?\n\nA configuration file would mean imposing the skeleton of the author\'s workflow\nonto all users. If such a workflow is suitable for your needs, example code\nand tutorial are provided so just have to adapt the code for yourself.\n\n## License\n\nThis project is distributed under the terms of the [WTFPL](./COPYING). It comes\nvoid of warranties and if you break things it\'s on you.\n',
    'author': 'Rémy Sanchez',
    'author_email': 'remy.sanchez@hyperthese.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/WithIO/luh3417',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from argparse import ArgumentParser
from typing import Optional, Sequence, Text

from luh3417.serialized_replace import walk
from luh3417.utils import make_doer, run_main, setup_logging

doing = make_doer("luh3417.replace")


def parse_args(argv: Optional[Sequence[Text]] = None):
    parser = ArgumentParser(description="Seeks and replaces serialized values")

    parser.add_argument("-i", "--input", required=True, help="Input file name")
    parser.add_argument("-o", "--output", required=True, help="Output file name")
    parser.add_argument("-b", "--before", nargs="+", help="String(s) to look for")
    parser.add_argument("-a", "--after", nargs="+", help="String(s) to replace by")
    parser.add_argument(
        "-c", "--charset", default="utf-8", help="What charset to use to read the file"
    )

    args = parser.parse_args(argv)

    if len(args.before) != len(args.after):
        parser.error("Not the same number of --before and --after")
        exit(1)

    return args


def main(argv: Optional[Sequence[Text]] = None):
    args = parse_args(argv)
    setup_logging()

    rep = [
        *zip(
            (x.encode(args.charset) for x in args.before),
            (x.encode(args.charset) for x in args.after),
        )
    ]

    with open(args.input, "rb") as i, open(args.output, "wb") as o:
        for line in i:
            o.write(walk(line, rep))


def __main__():
    return run_main(main, doing)


if __name__ == "__main__":
    __main__()



from copy import deepcopy
from typing import Dict, Text

from luh3417.luhfs import Location
from luh3417.utils import LuhError, import_file


class UnknownEnvironment(LuhError):
    pass


def load_generator(file_path: Text):
    """
    Loads the generator module from the specified file
    """

    return import_file("generator", file_path)


def apply_wp_config(patch: Dict, wp_config: Dict, source: Location) -> Dict:
    """
    Given a wp_config and a source, apply them into a patch

    - Make sure that the source arg is set
    - Put the DB settings in php_define
    - Also set wp_config properly
    """

    new_patch = deepcopy(patch)

    new_patch["args"] = {"source": f"{source}"}
    new_patch["wp_config"] = wp_config

    if "php_define" not in new_patch:
        new_patch["php_define"] = {}

    try:
        new_patch["php_define"].update(
            {
                "DB_HOST": wp_config["db_host"],
                "DB_USER": wp_config["db_user"],
                "DB_NAME": wp_config["db_name"],
                "DB_PASSWORD": wp_config["db_password"],
            }
        )
    except (KeyError, TypeError, AttributeError) as e:
        raise LuhError(f"Generated wp_config is incorrect: {e}")
    else:
        return new_patch
